// Copyright (C) 2010 Matthew Turnbull. All rights reserved.
// Use of this source code is governed by the 3-clause BSD License
// that can be found in the LICENSE file.

// This is stupid...
#define _GNU_SOURCE
#define _POSIX_SOURCE
#define _POSIX_C_SOURCE 200112

#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <signal.h>

#include "chatter.h"

static mesg_node_t* dispatch_list_head = NULL;
static mesg_node_t* dispatch_list_tail = NULL;
static pthread_mutex_t dispatch_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t dispatch_cond = PTHREAD_COND_INITIALIZER;

static client_sock_node_t* client_list_head = NULL;
static client_sock_node_t* client_list_tail = NULL;
static pthread_mutex_t client_mutex = PTHREAD_MUTEX_INITIALIZER;

// Read from the socket and attempt to fill the read buffer.
// Returns the number of bytes added (or -1 on failure).
int buffer_fill(client_sock_node_t* client)
{
	// Fail on buffer overflow
	// Also short curcuit fail if there is a marked error
	if(client->length == MAX_MESSAGE_LENGTH || client->error)
	{
		return -1;
	}

	// Maximize our usable available space
	if(client->start > 0)
	{
		if(client->length > 0)
		{
			memmove(client->buff, client->buff + client->start, client->length);
		}
		client->start = 0;
	}

	// Append new data to the buffer
	int max_read_len = MAX_MESSAGE_LENGTH - client->length;

	int rv;
	do
	{
		rv = read(client->sock, client->buff + client->length, max_read_len);
	}
	while(rv < 0 && errno == EINTR);

	if(rv < 0)
	{
		return -1;
	}
	client->length += rv;
	return rv;
}

// Parses the next \n delimited text line and allocates it to the buffer point to by **buff.
// Returns the lenfth of the string (including the \n) (or -1 on error).
int read_line(client_sock_node_t* client, char** buff)
{
	int rv;
	while(1)
	{
		// We've run out of buffer. Grab some more data
		if(client->index == client->length)
		{
			if((rv = buffer_fill(client)) <= 0)
			{
				client->index = 0;
				return ((rv == 0 && client->length == 0) ? 0 : -1);
			}
		}

		// Search for the next [\r]\n EOL
		while(client->index < client->length)
		{
			char c = client->buff[client->start + client->index];
			if(c == '\n')
			{
				// Strip [\r]\n from the line
				int len = client->index;
				if(len > 0 && client->buff[client->start + len - 1] == '\r')
				{
					len--;
				}

				// Allocate and move to buff
				*buff = (char*)malloc(sizeof(char) * (len + 2));
				if(!*buff)
				{
					client->index = 0;
					return -1;
				}
				memcpy(*buff, client->buff + client->start, len);
				(*buff)[len] = '\n';
				(*buff)[len + 1] = 0;

				client->start += client->index + 1;
				client->length -= client->index + 1;
				client->index = 0;

				return len + 1;
			}

			client->index++;
		}
	}
}

// Sends buff to the client. If there was an error, mark it to handle later.
// Returns the number of bytes sent (or -1 on error).
int send_line(client_sock_node_t* client, char* buff)
{
	char* buff_ptr = buff;
	int len = strlen(buff);
	int len_left = len;
	int rv;
	while(len_left)
	{
		if((rv = write(client->sock, buff_ptr, len_left)) < 0)
		{
			if(errno != EINTR)
			{
				// Mark there was an error so the client thread can handle it
				client->error = 1;
				return -1;
			}
			rv = 0;
		}
		len_left -= rv;
		buff_ptr += rv;
	}
	return len;
}


// Wrapper for inet_ntop. Returns the address (or dst on error).
const char* mynet_ntop(int af, const void *src, char *dst, int size)
{
	switch(af)
	{
		case AF_INET:
			return inet_ntop(af, &((struct sockaddr_in*)src)->sin_addr, dst, size);
		case AF_INET6:
			return inet_ntop(af, &((struct sockaddr_in6*)src)->sin6_addr, dst, size);
		default:
			return dst;
	}
}

// Wrapper for getting a port number from a sockaddr. Returns the port on success (or -1 on error).
int mynet_port_ntop(int af, const void *src)
{
	switch(af)
	{
		case AF_INET:
			return ntohs(((struct sockaddr_in*)src)->sin_port);
		case AF_INET6:
			return ntohs(((struct sockaddr_in6*)src)->sin6_port);
		default:
			return -1;
	}
}

// Constructs a client socket node. Returns the node (or NULL on error).
client_sock_node_t* new_client_sock_node(void)
{
	client_sock_node_t* client = (client_sock_node_t*)malloc(sizeof(client_sock_node_t));
	if(!client)
	{
		perror("Error on malloc");
		return NULL;
	}

	memset(client, 0, sizeof(client_sock_node_t));
	client->sock = -1;
	client->addr_len = sizeof(struct sockaddr);

	return client;
}

// Free a client socket node. Closes the connection if needed.
void free_client_sock_node(client_sock_node_t* node)
{
	if(node->sock >= 0)
	{
		close(node->sock);
	}
	free(node);
}

// Frees a client socket list.
void free_client_sock_list(client_sock_node_t* head)
{
	client_sock_node_t* next = head;
	client_sock_node_t* current;
	while(next)
	{
		current = next;
		next = current->next;
		free_client_sock_node(current);
	}
}

// Constructs a server socket node. Returns the node (or NULL on error).
srv_sock_node_t* new_serv_sock_node(void)
{
	srv_sock_node_t* server = (srv_sock_node_t*)malloc(sizeof(srv_sock_node_t));
	if(!server)
	{
		perror("Error on malloc");
		return NULL;
	}

	memset(server, 0, sizeof(srv_sock_node_t));
	server->sock = -1;
	server->addr_len = sizeof(struct sockaddr);

	return server;
}

// Free a serever socket node. Closes the connection if needed.
void free_serv_sock_node(srv_sock_node_t* node)
{
	if(node->sock >= 0)
	{
		close(node->sock);
	}
	free(node);
}

// Frees a lost of server sockets.
void free_serv_sock_list(srv_sock_node_t* head)
{
	srv_sock_node_t* next = head;
	srv_sock_node_t* current;
	while(next)
	{
		current = next;
		next = current->next;
		free_serv_sock_node(current);
	}
}

// Enqueues a message for dispatch. If this is the first message in the list, send a signal to
// The dispatch thread. Returns 0 on success (or -1 on error).
int message_enqueue(char* buff, client_sock_node_t* origin)
{
	mesg_node_t* mesg = (mesg_node_t*)malloc(sizeof(mesg_node_t));
	if(!mesg)
	{
		perror("Error on malloc");
		return -1;
	}
	mesg->buff = buff;
	mesg->origin = origin;
	mesg->next = NULL;

	pthread_mutex_lock(&dispatch_mutex);

	// Add it to the end of the dispatch list
	if(!dispatch_list_head)
	{
		dispatch_list_head = dispatch_list_tail = mesg;
		pthread_cond_signal(&dispatch_cond);
	}
	else
	{
		dispatch_list_tail = dispatch_list_tail->next = mesg;
	}

	pthread_mutex_unlock(&dispatch_mutex);

	return 0;
}

// Dequeues the next available message. If there is no message available, wait for a signal
// from a message enqueue. Returns the message (or NULL on failure).
mesg_node_t* message_dequeue(void)
{
	pthread_mutex_lock(&dispatch_mutex);

	if(!dispatch_list_head)
	{
		if(pthread_cond_wait(&dispatch_cond, &dispatch_mutex))
		{
			pthread_mutex_unlock(&dispatch_mutex);
			return NULL;
		}
	}

	mesg_node_t* mesg = dispatch_list_head;
	dispatch_list_head = dispatch_list_head->next;

	pthread_mutex_unlock(&dispatch_mutex);

	mesg->next = NULL;
	return mesg;
}

// The message dispatch thread. Dequeue a message and sent it to all of the clients.
void* message_dispatch_thread(void* args)
{
	mesg_node_t* mesg;
	while(1)
	{
		// Get the next message
		mesg = message_dequeue();
		if(!mesg)
		{
			continue; // this shouldn't happen
		}

		pthread_mutex_lock(&client_mutex);

		// Send it to the clients
		client_sock_node_t *current = client_list_head;
		while(current)
		{
			if(current != mesg->origin)
			{
				send_line(current, mesg->buff);
			}

			current = current->next;
		}

		pthread_mutex_unlock(&client_mutex);

		// Free the messages
		free(mesg->buff);
		free(mesg);
	}

	return NULL;
}

// Handle a client. Wait for a text line and enqueue it for the dispatch thread.
// On socket error, close it and remove it from the client list.
void* message_listen_thread(void* args)
{
	char address[INET6_ADDRSTRLEN];
	client_sock_node_t *client = (client_sock_node_t*)args;

	// Do the loop
	char* buff;
	int rv;
	while(!client->error && (rv = read_line(client, &buff)) > 0)
	{
		fprintf(stdout, "INFO:    Client message recvd %s:%i\n",
			mynet_ntop(client->addr_fam, &client->addr, address, client->addr_len),
			mynet_port_ntop(client->addr_fam, &client->addr));

		if((rv = message_enqueue(buff, client)) < 0)
		{
			break;
		}
	}

	fprintf(stdout, "INFO:    Client disconnect %s:%i (%s)\n",
		mynet_ntop(client->addr_fam, &client->addr, address, client->addr_len),
		mynet_port_ntop(client->addr_fam, &client->addr), ((rv == 0 && !client->error)
									? "client disconnect"
									: "socket/dispatch error"));

	pthread_mutex_lock(&client_mutex);

	// Remove the client
	client_sock_node_t *current, *last;
	current = client_list_head;
	while(current)
	{
		if(current == client)
		{
			if(current == client_list_head)
			{
				// Remove the first node
				client_list_head = current->next;
			}
			else
			{
				// Remove non-first node
				last->next = current->next;
			}

			// Update tail pointer
			if(current == client_list_tail)
			{
				client_list_tail = last;
			}

			break;
		}

		last = current;
		current = current->next;
	}

	pthread_mutex_unlock(&client_mutex);

	free_client_sock_node(client);

	return NULL;
}

// Accept a connectin from the connection. Returns CONNECT_OK on success.
// On error returns CONNECT_FATAL on general errors (malloc), CONNECT_SOCK
// on socket errors, or CONNECT_REFUSE if we're unable to handle a connection.
int accept_connection(srv_sock_node_t *conn)
{
	char address[INET6_ADDRSTRLEN];

	client_sock_node_t* client = new_client_sock_node();
	if(!client)
	{
		return CONNECT_FATAL;
	}

	// Grab the client
	client->sock = accept(conn->sock, &client->addr, &client->addr_len);
	if(client->sock < 0)
	{
		free_client_sock_node(client);
		return CONNECT_SOCK;
	}
	client->addr_fam = conn->addr_fam;

	fprintf(stdout, "INFO:    Client connect %s:%i\n",
		mynet_ntop(client->addr_fam, &client->addr, address, client->addr_len),
		mynet_port_ntop(client->addr_fam, &client->addr));

	pthread_mutex_lock(&client_mutex);

	// Create and run a detached thread
	pthread_attr_t attr;
	pthread_t pt;
	if(pthread_attr_init(&attr)
	|| pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)
	|| pthread_create(&pt, &attr, message_listen_thread, client))
	{
		pthread_mutex_unlock(&client_mutex);

		fprintf(stderr, "ERROR:   Unable to create message listen thread. Refusing connection.\n");
		free_client_sock_node(client);
		return CONNECT_REFUSE;
	}

	// Add it to the end of the client list
	if(!client_list_head)
	{
		client_list_head = client_list_tail = client;
	}
	else
	{
		client_list_tail = client_list_tail->next = client;
	}

	pthread_mutex_unlock(&client_mutex);

	return CONNECT_OK;
}

// Connect to the available configured network interfaces on port "port"
// Returns the head of a list of available sockets (or NULL on failure).
srv_sock_node_t* connect_srv_socks(char* port)
{
	// Validate port number
	int port_len = strlen(port);
	for(int i = 0; i < port_len; i++)
	{
		if(!isdigit(port[i]))
		{
			fprintf(stderr, "ERROR:   Invalid port. Must be an interger between 0 and 65535\n");
			return NULL;
		}
	}
	int port_int = atoi(port);
	if(port_int < 0 || port_int > 65535)
	{
		fprintf(stderr, "ERROR:   Invalid port. Must be an interger between 0 and 65535\n");
		return NULL;
	}

	// Generate our address hints
	struct addrinfo hints;
	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG; // | AI_V4MAPPED; // errors on grid machienes

	// Get the desired sockets' info
	struct addrinfo *addrs, *addr;
	int gai_err = getaddrinfo(NULL, port, &hints, &addrs);
	if(gai_err != 0)
	{
		fprintf(stderr, "ERROR:    getaddrinfo failed. %s\n", gai_strerror(gai_err));
		return NULL;
	}

	// Connect to the sockets
	srv_sock_node_t *sock_list = NULL, *sock_list_tail, *s;
	char address[INET6_ADDRSTRLEN];
	int on = 1;
	for(addr = addrs; addr; addr = addr->ai_next)
	{
		s = new_serv_sock_node();
		if(!s)
		{
			return NULL;
		}

		// Apparently some systems don't like using the same port for IPv4 and IPv6 addresses
		if(addr->ai_family == AF_INET6)
		{
			((struct sockaddr_in6*)addr->ai_addr)->sin6_port = htons(ntohs(((struct sockaddr_in6*)addr->ai_addr)->sin6_port) + 1);
		}

		// Open the socket
		s->sock = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);

		// If the socket is open, set SO_REUSEADDR. We don't really care if this fails
		if(s->sock >= 0)
		{
			setsockopt(s->sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int));
		}

		// Bind, and listen on the socket
		if(s->sock < 0
		|| bind(s->sock, addr->ai_addr, addr->ai_addrlen) < 0
		|| listen(s->sock, SOMAXCONN) < 0)
		{
			fprintf(stderr, "WARNING:    Unable to listen on %s:%i (%s)\n",
				mynet_ntop(addr->ai_family, &addr->ai_addr, address, addr->ai_addrlen),
				mynet_port_ntop(addr->ai_family, &addr->ai_addr),
				strerror(errno));
			free_serv_sock_node(s);
			continue;
		}
		s->addr = *(addr->ai_addr);
		s->addr_len = addr->ai_addrlen;
		s->addr_fam = addr->ai_family;

		fprintf(stdout, "INFO:    Listening on %s:%i\n",
			mynet_ntop(s->addr_fam, &s->addr, address, s->addr_len),
			mynet_port_ntop(s->addr_fam, &s->addr));

		// Add it to the list
		if(!sock_list)
		{
			sock_list = sock_list_tail = s;
		}
		else
		{
			sock_list_tail = sock_list_tail->next = s;
		}
	}
	freeaddrinfo(addrs);

	// Return the list of connected sockets
	return sock_list;
}

// Main entry point. Connect and listen on the server sockets,
// start the dispatch thread, and wait for incommin connections.
int main(int argc, char* argv[])
{
	// Print help output
	if(argc < 2 || strcasecmp(argv[1], "-h") == 0 || strcasecmp(argv[1], "-v") == 0)
	{
		fprintf(stdout, "Chat Relay Server (Chatter)\n");
		fprintf(stdout, "Release version 2010.08.13\n");
		fprintf(stdout, "\n");
		fprintf(stdout, "Copyright (C) 2010 Matthew Turnbull. All rights reserved.\n");
		fprintf(stdout, "Available under the 3-clause BSD License\n");
		fprintf(stdout, "<http://caligonstudios.com/files/programming/LICENSE>\n");

		if(argc < 2 || strcasecmp(argv[1], "-h") == 0)
		{
			char* prog_name = strrchr(argv[0], '/');
			if(!prog_name)
			{
				prog_name = argv[0];
			}
			else
			{
				prog_name++;
			}

			fprintf(stdout, "\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Usage: %s [OPTIONS] PORT\n", prog_name);
			fprintf(stdout, "Run the chat server on PORT (for IPv4) and PORT+1 (for IPv6) on all available network interfaces\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "PORT is the local port to run on (valid range 0 - 65535)\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Options:\n");
			fprintf(stdout, "   -h prints this usage message and exit\n");
			fprintf(stdout, "   -v prints version information and exit\n");
			fprintf(stdout, "\n");
			fprintf(stdout, "Examples:\n");
			fprintf(stdout, "   %s -h\n", prog_name);
			fprintf(stdout, "   %s 8080\n", prog_name);
		}

		exit((argc < 2) ? 1 : 0);
	}

	// Catch and ignore pipe errors
	if(signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	{
		fprintf(stderr, "WARNING:    Unable to block SIGPIPE. The server may unexpectedly terminate.\n");
	}

	// Run the message dispatch thread
	pthread_attr_t attr;
	pthread_t dispatch_thread;
	if(pthread_attr_init(&attr)
	|| pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)
	|| pthread_create(&dispatch_thread, &attr, message_dispatch_thread, NULL))
	{
		fprintf(stderr, "ERROR:   Unable to create message dispatch thread.\n");
		exit(1);
	}

	// Connect and listen on any available addresses
	srv_sock_node_t *sock_list = connect_srv_socks(argv[1]);
	if(!sock_list)
	{
		fprintf(stderr, "ERROR:   No available sockets for listening. Exiting.\n");
		exit(1);
	}

	// Wait for connections.
	fd_set inputfds;
	int max_sock;
	srv_sock_node_t *current, *last, *tmp;
	int rv;
	struct timeval timeout;
	char address[INET6_ADDRSTRLEN];
	while(sock_list) // && pthread_tryjoin_np(dispatch_thread, NULL) == EBUSY) // glib on grid machienes is too old :(
	{
		// Setup select
		FD_ZERO(&inputfds);
		max_sock = 0;
		for(current = sock_list; current; current = current->next)
		{
			FD_SET(current->sock, &inputfds);
			if(max_sock < current->sock)
			{
				max_sock = current->sock;
			}
		}
		timeout.tv_sec = 60;
		timeout.tv_usec = 0;

		// Do Select
		rv = select(max_sock + 1, &inputfds, NULL, NULL, &timeout);
		if(rv == 0)
		{
			continue;
		}
		else if(rv < 0)
		{
			perror("Error on select");
			fprintf(stdout, "INFO:   Closing server sockets.\n");
			free_serv_sock_list(sock_list);
			fprintf(stdout, "INFO:   Closing client connections.\n");
			pthread_mutex_lock(&client_mutex);
			free_client_sock_list(client_list_head);
			exit(1);
		}

		// Accept the available connections
		current = sock_list;
		last = NULL;
		while(current)
		{
			if(FD_ISSET(current->sock, &inputfds))
			{
				rv = accept_connection(current);
				// Ignore fatal (malloc) errors. The server may continue to operate.
				// If there is an error on the socket its self, remove it.
				if(rv == CONNECT_SOCK)
				{
					fprintf(stderr, "ERROR:    Error listening on %s:%i (removing)\n",
						mynet_ntop(current->addr_fam, &current->addr, address, current->addr_len),
						mynet_port_ntop(current->addr_fam, &current->addr));

					// Remove it from the server socket list
					if(current == sock_list)
					{
						sock_list = current->next;
					}
					else
					{
						last->next = current->next;
					}
					tmp = current;
					current = current->next;
					free_serv_sock_node(tmp);
					continue;
				}
			}

			last = current;
			current = current->next;
		}
	}

	return 0;
}

